package PureMessage::RPC::ClientV2;
use PureMessage::RPC::ResultV2;
use IO::Select;
use Data::Dumper;
use HTTP::Request::Common qw( POST $DYNAMIC_FILE_UPLOAD );
use MIME::Base64 qw(encode_base64 decode_base64);
use PureMessage::RPC::NamedDataEncoder;
use strict;

# Does the transport layer operations for RPC

our $DEBUG = $ENV{RPC_DEBUG};

$DYNAMIC_FILE_UPLOAD = 1;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my %self = ();
    $self{hosts} = shift;
    return unless ref($self{hosts}) eq "HASH";
    my %opts = @_;
    $self{timeout} = $opts{timeout} || 30; # TODO a configurable hook for timeout
    $self{socket} = {};
    $self{select} = {};
    $self{request_type} = "GET";
    $self{headers} = {Accept => 'text/csv_b64, text/csv, text/html'};
    return bless( \%self, $class);
}

sub event {
    # XXX use this to register custom callbacks
}

sub call {
    my $self = shift;
    my ($service, @args) = @_;
    $self->{attachments} = []; # clear any attachments 
    $self->{request_type} = "GET";
    $service =~ s/\./\//;
    # XXX raise_error if ! $service
    $self->{url} = "/RPC/$service/";
    if (scalar(@args)) {
      my $nd = PureMessage::RPC::NamedDataEncoder->new();
      $self->attach( DATA => $nd->create_named_line('arg',\@args));
      # for different types of data we could:
      #$self->attach( DATA => $nd->create_named_line('arg',$_) . "\n") foreach @args;
    }
    $self->{error} = {};
    $self->{select} = IO::Select->new();
    while ( my ($host, $info) = each %{$self->{hosts}} ) {
        # can we resolve this hostname?
        my $name = gethostbyname($info->{host});
        unless($name) {
           $self->raise_error($info->{host},"Unable to determine ip of rpc host '$info->{host}'");
           next;
        };
        $self->{result}{$host} = PureMessage::RPC::ResultV2->new();
        unless ( $info->{proto} ) {
            require PureMessage::ServerGroup;
            $info->{proto} = PureMessage::ServerGroup::proto($info);
        }
        my $socket;
        if ( $info->{proto} eq "HTTPS" ) {
            require Net::HTTPS::NB;
            $socket = Net::HTTPS::NB->new( Host => $info->{host} . ":" . $info->{port}, 
                                           SSL_verify_mode => 0 ) 
                                || ($self->raise_error($host, $@) and return);
        }
        elsif ( $info->{proto} eq "HTTP" ) {
            require Net::HTTP::NB;
            $socket = Net::HTTP::NB->new( Host => $info->{host} . ":" . $info->{port} )
                                || ($self->raise_error($host, $@) and return);
        }
        else {
            $self->{error}{$host} = "Unknown protocol '$info->{proto}' for RPC host '$info->{host}'";
            next;
        }
	if ( $socket ) {
            $self->{socket}{$socket} = $host;
            if ( $self->{request_type} eq "GET" ) {
                $socket->write_request($self->{request_type}, 
                        $self->{url},
                        %{$self->{headers}},
                        Cookie => $info->{auth} ? "sid=$info->{auth}, path=/" : "");
            }
            elsif ( $self->{request_type} eq "POST" ) {
                my $req = POST $self->{url},
                          Content_Type => 'form-data',
                          Content      => $self->{attachments};
                my $cont_ref = $req->content;
                $socket->write_request($self->{request_type}, 
                        $self->{url},
                        %{$self->{headers}},
                        %{$req->headers()},
                        Cookie => $info->{auth} ? "sid=$info->{auth}, path=/" : "");
                my $buf;
                # we only get a ref when there is a file attached
                if (ref($cont_ref) eq 'CODE') {
                    while( ($buf = &$cont_ref()), defined($buf) && length($buf)) {
                        $socket->print($buf);
                    }
                } else {
                    print "POSTED: $cont_ref\n" if $DEBUG;
                    $socket->print($cont_ref);
                }
            }
            else {
                die "Unsupported request type: $self->{request_type}";
            }
            $self->{select}->add($socket);
        }
        else {
            $self->{error}{$host} = "Unable to connected to RPC host '$host':$@";
        }
    }
}

sub wait {
    my $self = shift;
    print "TIMEOUT IS $self->{timeout}\n" if $DEBUG;
    print "NOW is: " . time . "\n" if $DEBUG;
    while ( my @ready = $self->{select}->can_read($self->{timeout}) ) {
        foreach my $socket ( @ready ) {
            my $host = $self->{socket}{$socket};
            next if $self->{result}{$host}{eof};
            if ( ! $self->{result}{$host}{code} ) {
                # expect headers
                my ($code, $mess, %h);
                eval {
                    ($code, $mess, %h) = $socket->read_response_headers();
                };
                if ( $@ ) {
                    $code = "501";
                    $mess = $!;
                    %h = ('X-Perl-Error' => $@);
                    print "Failure to read a http status code from server\n" if $DEBUG;
                } 
                if ($DEBUG) {
                  print "Header-Code: $code Message: $mess\n";
                  print "HEADER $_: $h{$_}\n" foreach keys %h;
                }
                next unless $code and $mess and %h;
                $self->{result}{$host}{code} = $code;
                $self->{result}{$host}{msg} = $mess;
                $self->{result}{$host}{headers} = \%h;
                if ( $code != 200 ) {
                    if ( $code == 403 ) {
                        $self->raise_error($host, "Access denied");
                    }
                    elsif ( $code == 401 ) {
                        $self->raise_error($host, "Authentication failed");
                    }
                }
            }
            else {
                my $buf;
                my $n = $socket->read_entity_body($buf, 1024);
                print "READ($n chars)\n" . as_hex($buf) if $DEBUG;
                $self->{error}{$host} = "Body read error" if ! defined $n;
                $self->{result}{$host}{eof} = 1 if defined $n and $n == 0;
                $self->{result}{$host}->parse($buf, 1);
                # XXX do data callback here
            }
        }
        my $status = 1;
        foreach my $host ( keys %{$self->{hosts}} ) {
            $status = 0 if not ( $self->{error}{$host} or $self->{result}{$host}{eof} );
        }
        last if $status;
    }
    foreach my $host (keys %{$self->{hosts}}) {
      $self->{error}{$host} = "Host '$host' never returned EOF. Possible timeout reading from it."
        if ! $self->{result}{$host}{eof} and !$self->{error}{$host};
    }

	$self->{select}->remove($self->{select}->handles);
	$self->{select} = {};
		
	foreach my $socket (keys(%{$self->{socket}})) {
		delete($self->{socket}->{$socket});
		close($socket);
	}

    print "NOW is: " . time . "\n" if $DEBUG;
}

sub as_hex {
  my $data = shift;
  my $string;
  my $substr;
  my $CHAR_BLOCK_SIZE = 16;
  my $FIELD_WIDTH = $CHAR_BLOCK_SIZE * 3;
  my $ddata = $data;
  return unless length($data) > 0;
  $ddata =~ s/^(.*)$/decode_base64($1)/gme; # base64 stops at first padding of =
  while($substr = substr($data,0,$CHAR_BLOCK_SIZE < length($data) ? $CHAR_BLOCK_SIZE : length($data),'')) {
    my $hex;
    $hex = join(' ',map { sprintf("%02x",ord($_)) } split(//,$substr));
    $substr =~ s/[[:cntrl:]]/./g;
    $string .= sprintf("%-${FIELD_WIDTH}s %s\n",$hex,$substr);
  }
  $string .= "\n";
  $substr = undef;
  while($substr = substr($ddata,0,$CHAR_BLOCK_SIZE < length($ddata) ? $CHAR_BLOCK_SIZE : length($ddata),'')) {
    my $hex;
    $hex = join(' ',map { sprintf("%02x",ord($_)) } split(//,$substr));
    $substr =~ s/[[:cntrl:]]/./g;
    $string .= sprintf("%-${FIELD_WIDTH}s %s\n",$hex,$substr);
  }
 
  return $string . "\n";
}


sub attach {
    # this is needed for Publishing API
    my $self = shift;
    push @{$self->{attachments}}, @_;
    $self->{request_type} = "POST";
}

sub data {
    my $self = shift;
    my $host = shift;
    if ( $host ) {
        return $self->{result}{$host}->data() if $self->{result}{$host};
    }
    else {
        return $self->{result};
    }
}

sub error {
    my $self = shift;
    my $host = shift;
    my @hosts = $host ? ($host) : keys %{$self->{hosts}};
    my @errors;
    foreach ( @hosts ) {
        my $err;
        if ( $err = $self->{error}{$_} ) {
            push @errors, scalar @hosts == 1 ? $err : "$_: " . $err;
        }
        elsif ( $self->{result}{$_} and $err = $self->{result}{$_}->error() ) {
            push @errors, scalar @hosts == 1 ? $err : "$_: " . $err;
        }
    }
    return 0 if scalar @errors < 1;
    return join("\n", @errors);
}

sub raise_error {
    my $self = shift;
    my $host = shift;
    $self->{error}{$host} = join(" ", @_);
}

sub urlize {
  my $rv = shift;
  $rv =~ s/([^A-Za-z0-9])/sprintf("%%%2.2X", ord($1))/ge;
  return $rv;
}

1;
