package PureMessage::RPC::NamedDataEncoder;
use PureMessage::MessageLog qw(split_line decode_value join_line);
use strict;
use warnings;

=head1 NAME

PureMessage::RPC::NamedDataEncoder - Prepare Marshalled Data

=head1 SYNOPSIS

# to encode data inside an rpc handler... 

my $eu = new PureMessage::RPC::NamedDataEncoder();

$self->dispatch( $eu->create_dispatch_record($name,@data)); 

# to recieve data in a client..
# 
my %data = ( host=>'localhost',proto=>'HTTP',prot=>28080);
my $b = PureMessage::RPC::ClientV2->new({ $host => \%data});

$b->call($RPC_METHOD,@arguments);
$b->wait();

if ($b->error()) {
    die $b->error();
} else {
   my @res;
   @res = $b->data($data{host});
   
   my $eu = new PureMessage::RPC::NamedDataEncoder();
   my ($results,$error) = $eu->parse_response_data(@res);
   if ($results) {
     if ($error) {
       foreach my $error_code (keys %$error) {
         print 'Error: ' . $error->{$error_code}->{errorstr};
       }
     }
     foreach $method_called (keys %$results) {
       my $data = $results->{method_called};
        # .. do stuff with the data
     }
     
   } else {
     # it's just normal rpc data, do whatever with it
   }
}
=head1 DESCRIPTION

This package can be used to encode and decode simple
perl data structures for transport over rpc.

=over

=cut

sub new {
  my $proto = shift;
  my %opts = @_;
  my $class = ref($proto) || $proto;
  my %self = ( );
  return bless(\%self,$class);
}

=item create_named_line($name,$data)

Will return an encoded string for the passed data.
Data can be a hash, array reference, or a scalar
value.  The returned string can be decoded later
with C<parse_named_line> to reconstruct the data
value and the name.

=cut

sub create_named_line {
  my ($self,$name,$data) = @_;
  my $line;
  if (ref($data) eq 'HASH') {
     $line = join_line("HASH{$name}",%$data);
  } elsif(ref($data) eq 'ARRAY') {
    my @format = map { + 'V' => $_ } @$data;
    $line = join_line("ARRAY{$name}",@format);
  } else {
    # it's a scalar
     $line = join_line("SCALAR{$name}",'V' => $data);
  }
  return $line;
}

=item create_dispatch_record($name[,\%data ,..])

Will return a data struct that can be dispatched using
the rpc dispatch method.

=cut

sub create_dispatch_record {
  my ($self,$name,@data) = @_;
  my @result;
  @data = ( { UNDEF => 1 } ) unless scalar(@data); # dispatch an undef
  push(@result,{ NAME=> $name, DATA => $self->create_named_line($name,$_) . "\n" }) foreach @data;
  return @result;
}

=item create_error_record($name[,\%data,..])

Will return a data struct that can be dispatched using
the rpc dispatch method and that will be recognized as
an error.

=cut


sub create_error_record {
  my ($self,$name,@data) = @_;
  my @result;
  foreach my $err (@data) {
    next unless ref($err) eq 'HASH';
    push(@result,{ NAME=> $name, DATA => $self->raise_error($name,%$err) . "\n" });
  }
  return @result;
}


=item raise_error($error_code[,%extra_data])

Will encode a specially formated named_line that will
be recognized as an error when decoded.

=cut

sub raise_error {
  my ($self,$error,%data) = @_;
  my $line = join_line("ERROR{$error}",name => $error, %data);
  return $line;
}

=item parse_response_data(@data)

Will parse an array of data of the format as returned by
C<PureMessage::RPC::ClientV2::call>. If the format of the
data is recognized it will be decoded and returned, otherwise
undef will be returned. The format of the result is an array 
of two elements, the first is the decoded data, the second 
any errors that were found in the data.

=cut

sub parse_response_data {
  my ($self,@in) = @_;
  if ($in[0] and ref($in[0]) eq 'HASH' and $in[0]->{DATA}) {
    # then it's of the format we expect
    my @new;
    my @data;
    push(@data,$_->{DATA}) foreach @in;
    return $self->parse_named_line(@data);
  } else {
    return;
  }
}

=item parse_named_line(@lines)

Will parse the passed lines and return the resulting data
and any errors that were found in the data. The result is
an array with two elements, the first array reference 
contains the data, the second the errors,if any, that were
found.

=cut

sub parse_named_line {
  my ($self,@in) = @_;
  my %result;
  my %error;
  foreach my $row (@in) {
    foreach my $line (split("\n",$row)) {
      my ($type_name,@values) = split_line($line);
      next unless $type_name =~ /^(\w+){(\w+)}$/;
      my ($type,$name) = ($1,$2);
      if($type eq 'HASH') {
        # although the marshalling code allows duplicate keys, we don't!
        my $r = { @values };
        # there is a special case where the RPC code uses a hash ref of
        # { UNDEF => 1} as a sentinel to mean that a call to a method
        # returned no data. We need to catch that and make an empty
        # hash of it
        $r = {} if exists($r->{UNDEF}) and $r->{UNDEF} and scalar(keys(%$r)) == 1;
        if (exists($result{$name})) {
          $result{$name} = [$result{$name}] unless ref($result{$name}) eq 'ARRAY';
          push(@{$result{$name}},$r);
        } else {
          $result{$name}= $r;
        }
      } elsif ($type eq 'ARRAY') {
        my @V = ();
        while( my($k,$v) = splice(@values,0,2)) {
          next unless $k eq 'V'; # some kind of error if it's not our 'V' marker!
          push(@V,$v);  
        }
        if (exists($result{$name})) {
          $result{$name} = [$result{$name}] unless ref($result{$name}) eq 'ARRAY';
          push(@{$result{$name}},@V);
        } else {
          $result{$name} = \@V;
        }

      } elsif ($type eq 'SCALAR') {
        if (exists($result{$name})) {
          $result{$name} = [$result{$name}] unless ref($result{$name}) eq 'ARRAY';
          push(@{$result{$name}},$values[1]);
        } else {
          $result{$name} = $values[1];
        }
      } elsif ($type eq 'ERROR') {
        $error{$name} = { @values };
      } else {
        # no idea what it is
      }
    }
  }
  return (\%result,\%error); 
}

=pod

=back

=cut


1;

