# provides data serialization functionality
package PureMessage::RPC::Response;
use Text::CSV_XS;
use MIME::Base64 qw(encode_base64 decode_base64);

our $USE_BASE64_ENCODED = 1;
{ 
    no warnings qw(once);
    our $DEBUG = $ENV{RPC_DEBUG};
    our $DEBUG_SERVE = $ENV{RPC_DEBUG_SERVE};
};

=head1 NAME

PureMessage::RPC::Response - base object for rpc handlers

=head1 SYNOPSIS


  my $resp = new PureMessage::RPC::Response(_request => $request_obj, buffer=>1);


  $resp->dispatch("this is my response");
  $resp->commit(); # only needed when buffered

=head1 DESCRIPTION


This base class should never need to be called directory, but should only be
called by the C<RPC::Request> package ... 

todo

=over 4

=cut


=item new(%options)

Creates an response object. Required keys are:
_request => reference to the Request object that created this Response
buffer => boolean value that indicates wether this object should buffer stdout or not

=cut

sub new {
    my $proto = shift;
    my %options = @_;
    my $class = ref($proto) || $proto;
    if($USE_BASE64_ENCODED and (!exists($options{base64}) or $options{base64})) {
      $options{headers} = { 'Content-Type' => 'text/csv_b64' };
    } else {
      $options{headers} = { 'Content-Type' => 'text/csv' };
    }
    $options{serializer} = Text::CSV_XS->new({binary=>1});
    return bless( \%options, $class);
}

sub raise_error {
    my $self = shift;
    my $err_msg = shift;
    $self->dispatch( { ERROR => $err_msg });
    $self->_dispatched_errors();
}
sub _dispatched_errors {
    my $self = shift;
    $self->{error_count} = defined($self->{error_count}) ? $self->{error_count}++ : 1;
}

sub dispatched_errors {
    my $self = shift;
    return $self->{error_count};
}

sub dispatch_headers {
    my $self = shift;
    my $sample = shift;
    while ( my ($k, $v) = each %{$self->{headers}} ) {
        $self->serve("$k: $v\r\n");
    }
   # end of HTTP headers
    $self->serve("\r\n");
    
    # print column names if needed
    if ( ! $self->{meta_data} and $sample ) {
        if ( ref($sample) eq "HASH" ) {
            $self->{meta_data} = [ keys %{$sample} ];
        } elsif ( ref($sample) eq "ARRAY" ) {
            $self->{meta_data} = ['ARRAY'];
        } else {
            $self->{meta_data} = ['DATA'];
        }
    }
    $self->serve($self->serialize(@{$self->{meta_data}})) if $self->{meta_data};
    $self->{done_headers} = 1;
}

sub dispatch_raw {
    # dispatch preserialized data
    my $self = shift;
    if ( ! $self->{done_headers} ) {
        $self->dispatch_headers();
    }
    foreach ( @_ ) {
        $self->serve($_);
    }
}

sub dispatch {
    my $self = shift;
    my @data = @_;
    if ( ! $self->{done_headers} ) {
        $self->dispatch_headers($data[0]);
    }
    foreach my $record ( @data ) {
        if ( ref($record) eq "HASH" ) {
            $self->serve( $self->serialize( map {$record->{$_}} @{$self->{meta_data}} ) );
        }
        elsif ( ref($record) eq "ARRAY" ) {
            $self->serve($self->serialize(@{$record}));
        }
        else {
            $self->serve($self->serialize($record));
        }
    }
}

sub serialize {
    my $self = shift;
    my @data = @_;
   
    if ($USE_BASE64_ENCODED) {
      @data = map { encode_base64($_,'') } @data;
    }
    
    if ( $self->{serializer}->combine(@data) ) {
        return $self->{serializer}->string()."\n";
    }
    else {
	# XXX should do better handling for "run-time" errors
        die "ERROR\nData serialization error";
    }
}

sub commit {
  my $self = shift;
  foreach my $row (@{$self->{_row_cache}}) {
    print $row;
  }
  printf("Row cache was commited at %s,%s:%s\n",caller(1)) if $DEBUG;
  $self->{_row_cache} = undef;
}

# only for testing
sub _get_buffer {
  my $self = shift;
  my $text;
  foreach my $row (@{$self->{_row_cache}}) {
    $text .= "$row"; 
  }
  $self->{_row_cache} = undef;
  $text;
}

sub serve {
    my $self = shift;
    if ($self->{buffer}) {
      if ($DEBUG_SERVE) {
          # then tag every line with were it came from in the code
          my @w;
          for(my $depth = 3; $depth >= 1; $depth--) {
            my @c = caller($depth);
            push(@w,join(',',@c[0..3]));
          }
          push(@{$self->{_row_cache}},join("\n",@w) . "\n[$_]\n") foreach (@_); 
      } else {
        push(@{$self->{_row_cache}},@_);
      }
    } else {
      print $_ foreach @_;
    }
}

# default method
sub doc {
    my $self = shift;
    $self->dispatch("PureMessage RPC Framework");
}

1;
